import os
import sys
import logging


class Logger:
    def __init__(self, log_to='both', log_dir='.', run_name='run'):
        self.log_to = log_to
        self.log_dir = log_dir
        self.run_name = run_name
        
        self.logger = logging.getLogger('kerosene')
        self.logger.setLevel(logging.DEBUG)
        self.logger.handlers.clear()
        self.logger.propagate = False

        if self.log_to == 'both' or self.log_to == 'file':
            self.set_file_logger()
        if self.log_to == 'both' or self.log_to == 'console':
            self.set_stream_logger()

    def set_stream_logger(self):
        sh = logging.StreamHandler(sys.stdout)
        sh.setFormatter(logging.Formatter('%(message)s'))
        self.logger.addHandler(sh)

    def set_file_logger(self):
        os.makedirs(self.log_dir, exist_ok=True)
        fh = logging.FileHandler(os.path.join(self.log_dir, self.run_name + '.log'))
        fh.setFormatter(logging.Formatter('%(asctime)s: %(message)s'))
        self.logger.addHandler(fh)

    def epoch(self, epoch, metrics):
        values = []
        for phase in metrics.keys():
            for metric, value in metrics[phase].items():
                values.append(f'{phase}_{metric}: {value:.5f}')
        values = ', '.join(values)
        self.logger.info(f'Epoch {epoch}: {values}')

    def test(self, metrics):
        log_text = [f'{metric}: {value:.5f}' for metric, value in metrics.items()]
        log_text = ', '.join(log_text)
        self.logger.info(f'Test results (avg): {log_text}')

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)

    def exception(self, message):
        self.logger.exception(message)



import torch


def to_device(x, device):
    if torch.is_tensor(x):
        return x.to(device)
    elif isinstance(x, list) or isinstance(x, tuple):
        x_type = type(x)
        res = [to_device(item, device) for item in x]
        return x_type(res)
    elif isinstance(x, dict):
        res = {k: to_device(v, device) for k, v in x.items()}
        return res
    return x



import os
from datetime import datetime
import torch
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from .utils.logger import Logger
from .utils.metrics import Metrics
from .utils.utils import to_device


class Trainer:
    def __init__(
        self,
        model,
        optimizer,
        data_loaders,
        criterion=None,
        scheduler=None,
        use_mixed_precision=True,
        model_name=None,
        run_name=None,
        device=None,
        output_dir='output',
        save_checkpoints='last',
        checkpoints_dir='checkpoints',
        use_tqdm=True,
        log_to='both',
        log_dir='.',
        use_tensorboard=True,
        tensorboard_dir='runs',
        save_metrics=True,
        save_curves=True
    ):
        if device is None:
            device = torch.device('cuda:0') if torch.cuda.is_available() else 'cpu'

        self.model_name = model_name or model.__class__.__name__
        self.run_name = run_name or self.model_name+'_'+datetime.now().strftime('%b-%d_%H-%M')

        # Training
        self.model = model.to(device)
        self.optimizer = optimizer
        self.criterion = criterion
        self.scheduler = scheduler
        self.use_mixed_precision = use_mixed_precision
        self.data_loaders = data_loaders
        self.device = device
        self.output_dir = output_dir
        self.save_metrics = save_metrics
        self.save_curves = save_curves

        # Checkpoints
        self.save_checkpoints = save_checkpoints
        self.checkpoints_dir = os.path.join(self.output_dir, checkpoints_dir)

        # Logging
        self.log_dir = os.path.join(self.output_dir, log_dir)
        self.logger = Logger(log_to=log_to, log_dir=self.log_dir, run_name=self.run_name)
        self.use_tqdm = use_tqdm

        # Tensorbaord
        self.use_tensorboard = use_tensorboard
        self.tensorboard_dir = os.path.join(self.output_dir, tensorboard_dir)
        if self.use_tensorboard:
            tb_logdir = os.path.join(self.tensorboard_dir, self.run_name)
            self.tb_writer = SummaryWriter(tb_logdir)

    # HOOKS

    def before_train(self, *args, **kwargs):
        pass

    def after_train(self, *args, **kwargs):
        pass

    def before_epoch(self, *args, **kwargs):
        pass

    def after_epoch(self, *args, **kwargs):
        pass

    def before_test(self, *args, **kwargs):
        pass

    def after_test(self, *args, **kwargs):
        pass

    # CHECKPOINTS

    def save_checkpoint(self, epoch):
        os.makedirs(self.checkpoints_dir, exist_ok=True)

        if self.save_checkpoints.startswith('every:'):
            every = int(self.save_checkpoints.split(':')[1])
            if epoch % every == 0:
                filename = f'{self.run_name}_{epoch}.pt'
            else:
                filename = None
        elif self.save_checkpoints == 'all':
            filename = f'{self.run_name}_{epoch}.pt'
        elif self.save_checkpoints == 'last':
            filename = f'{self.run_name}.pt'
        else:
            filename = f'{self.run_name}.pt'
        
        if filename is not None:
            checkpoint_filepath = os.path.join(self.checkpoints_dir, filename)
            model_dict = {
                'model_state_dict': self.model.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'epoch': epoch
            }
            torch.save(model_dict, checkpoint_filepath)
            # self.logger.info(f'Checkpoint saved at {checkpoint_filepath}.')

    def load_checkpoint(self, run_name):
        if os.path.isfile(run_name):
            checkpoint_filepath = run_name
        else:
            checkpoint_filepath = os.path.join(self.checkpoints_dir,  f'{run_name}.pt')

        checkpoint = torch.load(checkpoint_filepath)
        self.model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        self.logger.info(f'Loaded {run_name} checkpoint at epoch {checkpoint["epoch"]}.')

    # TRAINING

    def make_pbar(self, iterable):
        return tqdm(iterable, leave=False, disable=not self.use_tqdm)

    def training_step(self, batch, batch_idx):
        raise NotImplementedError

    def validation_step(self, batch, batch_idx):
        raise NotImplementedError

    def test_step(self, batch, batch_idx):
        raise NotImplementedError

    def train(self, epochs):
        self.before_train()
        self.logger.info(f'Starting run {self.run_name}: {self.model_name} with {epochs} epochs in {self.device}...')
        phases = list(self.data_loaders.keys() - {'test'})
        training_metrics = Metrics()

        scaler = torch.cuda.amp.GradScaler(enabled=self.use_mixed_precision)

        for epoch in range(epochs):
            self.before_epoch(epoch=epoch)

            for phase in phases:
                # Training phase
                if phase == 'train':
                    self.model.train()
                    pbar = self.make_pbar(self.data_loaders[phase])
                    for batch_idx, batch in enumerate(pbar):
                        batch = to_device(batch, self.device)
                        self.optimizer.zero_grad()

                        with torch.cuda.amp.autocast(dtype=torch.float16, enabled=self.use_mixed_precision):
                            metrics = self.training_step(batch, batch_idx)
                            loss = metrics['loss']

                        scaler.scale(loss).backward()
                        scaler.step(self.optimizer)
                        scaler.update()
                        pbar.set_description(f'[{epoch+1}/{epochs}] loss: {loss.item():.3f}')
                        training_metrics.add_batch_metrics(phase, metrics)
                # Validation phase
                else:
                    self.model.eval()
                    with torch.no_grad():
                        pbar = self.make_pbar(self.data_loaders[phase])
                        for batch_idx, batch in enumerate(pbar):
                            batch = to_device(batch, self.device)
                            metrics = self.validation_step(batch, batch_idx)
                            loss = metrics['loss']
                            pbar.set_description(f'[{epoch+1}/{epochs}] loss: {loss.item():.3f}')
                            training_metrics.add_batch_metrics(phase, metrics)

                epoch_metrics = training_metrics.aggretate_batches()

                # Apply schedular
                if self.scheduler is not None and phase == 'train':
                    if isinstance(self.scheduler, list) or isinstance(self.scheduler, tuple):
                        for scheduler in self.scheduler:
                            scheduler.step()
                    else:
                        self.scheduler.step()

                # Log metrics in Tensorboard
                if self.use_tensorboard:
                    for metric in epoch_metrics[phase]:
                        self.tb_writer.add_scalar(f'{phase}/{metric}', epoch_metrics[phase][metric], epoch)
                    
                    for name, weight in self.model.named_parameters():
                        self.tb_writer.add_histogram(name, weight, epoch)

            self.logger.epoch(epoch, epoch_metrics)

            if self.save_checkpoints is not None:
                self.save_checkpoint(epoch+1)
            
            self.after_epoch(epoch=epoch, metrics=epoch_metrics)

        self.logger.info('Finished training.')

        if self.use_tensorboard:
            self.tb_writer.close()
        
        # Save training metrics
        if self.save_metrics:
            training_metrics.save_json(os.path.join(self.output_dir, f'{self.run_name}.json'))

        # Save training curves
        if self.save_curves:
            training_metrics.save_plot(os.path.join(self.output_dir, f'{self.run_name}.png'))

        self.after_train(metrics=training_metrics)

        return training_metrics.values

    
    def test(self, test_data_loader=None, reduction='none'):
        self.before_test()
        self.logger.info('Starting test...')
        test_metrics = Metrics()

        if test_data_loader is None:
            test_data_loader = self.data_loaders['test']

        self.model.eval()
        with torch.no_grad():
            pbar = self.make_pbar(test_data_loader)
            for batch_idx, batch in enumerate(pbar):
                batch = to_device(batch, self.device)
                metrics = self.test_step(batch, batch_idx)
                test_metrics.add_batch_metrics('test', metrics)
            test_metrics.aggretate_batches(aggretation='concat')

        reduced_output = test_metrics.reduce(reduction)['test']

        log_metrics = reduced_output if reduction == 'mean' else test_metrics.reduce('mean')['test']
        self.logger.test(log_metrics)

        self.after_test()

        return reduced_output



from setuptools import setup, find_packages


setup(
    name='pytorch-kerosene',
    version='0.2.2',
    author='Silvan Ferreira',
    author_email='silvanfj@gmail.com',
    description='PyTorch Kerosene is a tool for abstracting the training process of PyTorch models.',
    packages=find_packages(),
    keywords=['python', 'pytorch'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import json
import numpy as np
import torch
import matplotlib.pyplot as plt


class Metrics:
    def __init__(self):
        self.batch_values = {}
        self.values = {}

    def add_batch_metrics(self, phase, metrics):
        for metric, value in metrics.items():
            if isinstance(value, torch.Tensor):
                metrics[metric] = value.detach().cpu()
                
                if value.dim() == 0:
                    metrics[metric] = metrics[metric].item()
                else:
                    metrics[metric] = metrics[metric].tolist()

        if phase not in self.batch_values:
            self.batch_values[phase] = {}
        
        for metric, value in metrics.items():
            if metric not in self.batch_values[phase]:
                self.batch_values[phase][metric] = []
            self.batch_values[phase][metric].append(value)

        return self.batch_values

    def aggretate_batches(self, aggretation='mean'):
        epoch_metrics = self.batch_values.copy()
        self.batch_values = {}
        for phase in epoch_metrics.keys():
            for metric, values in epoch_metrics[phase].items():
                if phase not in self.values:
                    self.values[phase] = {}
                if metric not in self.values[phase]:
                    self.values[phase][metric] = []

                # Flatten list of lists
                if isinstance(values[0], list):
                    values = [item for sublist in values for item in sublist]

                aggretation_funcs = {
                    'mean': np.mean,
                    'median': np.median,
                    'sum': np.sum,
                    'max': np.max,
                    'min': np.min,
                    'concat': lambda x: x
                }

                agg_func = aggretation_funcs[aggretation]
                epoch_metrics[phase][metric] = agg_func(values)

                if aggretation == 'concat':
                    self.values[phase][metric] += epoch_metrics[phase][metric]
                else:
                    self.values[phase][metric].append(epoch_metrics[phase][metric])

        return epoch_metrics
    
    def save_json(self, path):
        with open(path, 'w') as f:
            json.dump(self.values, f, indent=4)

    def save_plot(self, path):
        phases = list(self.values.keys())
        metrics = list(self.values[phases[0]].keys())
        epochs = range(1, len(self.values[phases[0]][metrics[0]]) + 1)

        ncols = 2
        nrows = (len(metrics) + ncols - 1) // ncols

        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(12, 4*nrows))

        for i, metric in enumerate(metrics):
            row = i // ncols
            col = i % ncols
            ax = axes[row, col] if nrows > 1 else axes[col]

            for phase in phases:
                ax.plot(epochs, self.values[phase][metric], label=f'{phase} {metric}')

            ax.xaxis.set_major_locator(plt.MaxNLocator(integer=True))
            ax.set_xlabel('epoch')
            ax.set_ylabel(metric)
            ax.set_title(f'{metric} over time')
            ax.grid('on', linestyle='--', linewidth=0.5, color='black', alpha=0.3)
            ax.legend()
        
        plt.tight_layout()
        plt.savefig(path)

    def reduce(self, reduction='mean'):
        reduction_funcs = {
            'mean': np.mean,
            'median': np.median,
            'min': np.min,
            'max': np.max,
            'none': lambda x: x
        }

        if isinstance(reduction, str):
            reduction_func = reduction_funcs[reduction]
        elif isinstance(reduction, function):
            reduction_func = reduction
        else:
            reduction_func = reduction_funcs['none']

        reduced_output = {}
        for phase, metrics in self.values.items():
            reduced_output[phase] = {}
            for metric, values in metrics.items():
                reduced_output[phase][metric] = reduction_func(values)

        return reduced_output

    def __getitem__(self, keys):
        if isinstance(keys, str):
            keys = (keys,)
        values = self.values
        for key in keys:
            values = values[key]
        return values
    
    def __repr__(self):
        return str(self.values)
